/*
 * mpeg2cleaner - Entfernt 'zero byte stuffing'-Bereiche aus MPEG2-ES-Dateien.
 * Homepage: http://basicmaster.de/mpeg2cleaner/
 * 
 * Copyright 2011-2015 Stefan Pöschel
 * Das Programm darf kostenlos zu privaten, nichtkommerziellen Zwecken benutzt werden (Benutzung auf eigene Gefahr).
 */

#include "mpeg2cleaner.h"


// --- sStatistics -----------------------------------------------------------------
inline void sStatistics::ApplyRegion(int fillbytes_region) {
	count++;
	fillbytes_total += fillbytes_region;
	if(fillbytes_region > fillbytes_max)
		fillbytes_max = fillbytes_region;
	if(fillbytes_region < fillbytes_min)
		fillbytes_min = fillbytes_region;
}


// --- cMPEG2Cleaner -----------------------------------------------------------------
cMPEG2Cleaner::cMPEG2Cleaner(const char *source_filename, const char *target_filename, const bool mode_verbose, const bool mode_write) {
	this->source_filename = source_filename;
	this->target_filename = target_filename;
	this->mode_verbose = mode_verbose;
	this->mode_write = mode_write;

	source_len = 0;

	source_buffer = new uint8_t[SOURCE_BUFFER_SIZE];
	source_buffer_level = 0;

	curr_slice = 0;
	rb_offset = 0;
	data_offset = 0;
	fillbytes_region = 0;
	fillbytes_pic = -1;

	progress = -1;
	filler_percentage = 0;
	timestamp_start = 0;
}

cMPEG2Cleaner::~cMPEG2Cleaner() {
	delete[] source_buffer;
}

inline void cMPEG2Cleaner::FillSourceBuffer() {

	// determine available bytes
	source_buffer_level = source_len - source_file.tellg();
	if(source_buffer_level > SOURCE_BUFFER_SIZE)
		source_buffer_level = SOURCE_BUFFER_SIZE;

	// fill buffer
	source_file.read((char*) source_buffer, source_buffer_level);
	rb_offset = 0;
}

inline void cMPEG2Cleaner::UpdateProgress(const bool initial_call) {
	uint64_t source_pos = source_file.tellg();

	// determine new progress - abort, if no change
	int new_progress = floor(100.0 * (double) source_pos / (double) source_len);
	if(progress == new_progress)
		return;

	// update progress
	progress = new_progress;

	int removed_bytes = slice_stat[0].fillbytes_total;
	double source_rate = 0;

	// if not initial call, update progress statistics
	if(!initial_call) {
		filler_percentage = (double) removed_bytes / (double) source_pos * 100.0;
		source_rate = (double) source_pos / CurrentDurationSec() / 1024.0 / 1024.0;
	}

	// display progress (CR beforehand)
	cout << "\rFortschritt: " << setw(3) << progress << "% / Bytes entfernt: " << setw(11) << removed_bytes << " (" << setw(3) << filler_percentage << "%) / Leserate: " << setw(5) << source_rate << " MB/s" << flush;
}

inline void cMPEG2Cleaner::FillerEnd() {
	// update slice/pic stats

	slice_stat[0].ApplyRegion(fillbytes_region);
	slice_stat[curr_slice].ApplyRegion(fillbytes_region);

	if(fillbytes_pic != -1)
		fillbytes_pic += fillbytes_region;
}

inline void cMPEG2Cleaner::PicBoundary() {
	if(fillbytes_pic != -1)
		pic_stat.ApplyRegion(fillbytes_pic);
	fillbytes_pic = 0;
}

inline void cMPEG2Cleaner::DataBoundary(const bool data_continues) {
	if(data_offset != -1) {
		// write data to target file
		if(mode_write)
			target_file.write((char*) source_buffer + data_offset, rb_offset - data_offset);
		data_offset = data_continues ? 0 : -1;
	}
}

inline void cMPEG2Cleaner::RemoveZeroByteStuffing() {
	int zero_count = 0;
	bool prefix_seen = false;
	bool in_slice = false;
	bool in_filler = false;
	

	printf("Verarbeite: %s\n", source_filename);
	if(mode_write)
		printf("Ausgabe in: %s\n", target_filename);
	printf("\n");
	
	// open source file
	source_file.open(source_filename, ios::binary);
	if(!source_file.is_open()) {
		printf("Fehler: Fehler beim Öffnen der Quelldatei!\n");
		exit(EXIT_FAILURE);
	}
	
	source_file.seekg(0, ios_base::end);
	source_len = source_file.tellg();
	source_file.seekg(0, ios_base::beg);

	FillSourceBuffer();

	// ensure present MPEG-2 sequence header
	if(source_buffer_level < 4 || source_buffer[0] != 0x00 || source_buffer[1] != 0x00 || source_buffer[2] != 0x01 || source_buffer[3] != 0xB3) {
		printf("Fehler: Quelldatei beginnt nicht mit MPEG-2 Sequence Header Code!\n");
		source_file.close();
		exit(EXIT_FAILURE);
	}

	// open target file - overwriting existing file
	if(mode_write)
		target_file.open(target_filename, ios::binary | ios::trunc);

	// init output format
	cout.precision(1);
	cout.setf(ios::fixed, ios::floatfield);
	cout.imbue(locale(cout.getloc(), new german_numpunct()));

	timestamp_start = clock();
	UpdateProgress(true);

	// process source buffer
	while(source_buffer_level) {
		uint8_t rb = source_buffer[rb_offset];
		bool in_filler_next = in_filler;

		// start code (prefix) detection
		if(prefix_seen) {
			prefix_seen = false;

			// check for slice
			in_slice = rb >= 0x01 && rb <= SLICE_MAX;
			if(in_slice)
				curr_slice = rb;

			// check for picture start
			if(rb == 0x00)
				PicBoundary();
		} else {
			if(rb == 0x00) {
				// if not in filler mode, in slice and enough zero bytes, change to filler mode
				if(!in_filler && in_slice && zero_count >= 3)		// 3 = 1 for possible padding of previous syntactic block + 2 for start of next start code prefix
					in_filler_next = true;

				zero_count++;
			} else {
				if(zero_count >= 2 && rb == 0x01)
					prefix_seen = true;
				in_filler_next = false;	// leave filler mode

				zero_count = 0;
			}
		}

		// execute transition + apply next state
		if(!in_filler && in_filler_next) {
			FillerStart();
			DataBoundary(false);
		}
		if(in_filler && !in_filler_next) {
			FillerEnd();
			DataStart();
		}
		in_filler = in_filler_next;

		// update counters
		if(in_filler)
			fillbytes_region++;
		rb_offset++;
		
		// if end of source buffer reached, try to refill it
		if(rb_offset == source_buffer_level) {
			DataBoundary(true);
			UpdateProgress(false);
			FillSourceBuffer();
		}
	}
	
	// tidy up stats/data
	if(in_filler)
		FillerEnd();
	PicBoundary();
	DataBoundary(false);

	// close files
	if(mode_write)
		target_file.close();
	source_file.close();

	// output summary (mixing cout and printf is bad style, i know)
	printf("\n");
	cout << "Dateigroesse: " << source_len << " Bytes => " << (source_len - slice_stat[0].fillbytes_total) << " Bytes / Dauer: " << CurrentDurationSec() << " s\n" << flush;
	
	if(!mode_verbose)
		return;

	// output verbose stats

	// only display caption, if filler bytes were removed at all
	if(slice_stat[0].fillbytes_total) {
		printf("\n");
		printf("Slice|Bildzeilen|Gesamt Bytes|  Min. Bytes|  Max. Bytes|  # Bereiche\n");
	}

	// handle slices 0x01 to SLICE_MAX, then total
	uint8_t prev_slice = 0;
	for(int i = 1; i <= SLICE_MAX + 1; i++) {
		curr_slice = i % (SLICE_MAX + 1);
		
		// skip slice, if no filler bytes were removed
		if(!slice_stat[curr_slice].fillbytes_total)
			continue;

		// insert empty/separation line, if the previous slice (having filler bytes removed) is not adjacent
		if(prev_slice != curr_slice - 1)
			printf("%s\n", curr_slice ? "" : "____________________________________________________________________");

		if(curr_slice)
			printf(" 0x%02X  %4d-%4d", curr_slice, curr_slice * 16 - 15, curr_slice * 16);
		else
			cout << "Gesamt: " << setw(5) << filler_percentage << "% =";
		cout << " " << setw(12) << slice_stat[curr_slice].fillbytes_total << " " << setw(12) << slice_stat[curr_slice].fillbytes_min << " "<< setw(12) << slice_stat[curr_slice].fillbytes_max << " " << setw(12) << slice_stat[curr_slice].count << "\n" << flush;

		prev_slice = curr_slice;
	}
	
	// pic stats
	printf("\n");
	cout << "Picture:                      " << setw(12) << pic_stat.fillbytes_min << " " << setw(12) << pic_stat.fillbytes_max << "\n" << flush;
}


int main(int argc, char** argv) {
	printf("mpeg2cleaner " MPEG2CLEANER_VERSION " - Copyright Stefan Poeschel " MPEG2CLEANER_YEAR "\n");
	printf("\n");

	// if no params, display usage
	if(argc == 1) {
		printf("Entfernt 'zero byte stuffing'-Bereiche aus MPEG-2-ES-Dateien.\n");
		printf("\n");
		printf("Syntax: %s [PARAMETER] QUELLDATEI [ZIELDATEI]\n", argv[0]);
		printf("\n");
		printf("  QUELLDATEI = Quelldateiname\n");
		printf("  ZIELDATEI  = Zieldateiname (Vorgabe: QUELLDATEI mit Endung '.CLEANED.m2v')\n");
		printf("  -v         = Ausgabe von Statistiken am Ende\n");
		printf("  -a         = nur Analyse (keine Zieldatei schreiben)\n");
		return EXIT_FAILURE;
	}

	string source_filename = "";
	string target_filename = "";
	bool mode_verbose = false;
	bool mode_write = true;

	// parse params
	for(int i = 1; i < argc; i++) {
		string param = argv[i];
		
		if(param == "-v") {
			mode_verbose = true;
			continue;
		}
		if(param == "-a") {
			mode_write = false;
			continue;
		}
		if(source_filename == "")
			source_filename = param;
		else
			target_filename = param;
	}

	if(source_filename == "") {
		printf("Fehler: Quelldateiname fehlt!\n");
		return EXIT_FAILURE;
	}
	
	// if not specified, assemble target filename
	if(target_filename == "") {
		target_filename = source_filename;

		if(target_filename.size() >= 4 && target_filename.substr(target_filename.size() - 4, 4) == ".m2v")
			target_filename.insert(target_filename.size() - 4, ".CLEANED");
		else
			target_filename += ".CLEANED.m2v";
	}

	cMPEG2Cleaner *m2c = new cMPEG2Cleaner(source_filename.c_str(), target_filename.c_str(), mode_verbose, mode_write);
	m2c->RemoveZeroByteStuffing();
	delete m2c;

	return EXIT_SUCCESS;
}
